%%% Filename : ac_analytical.pl
%%% Date     : 23.02.2011
%%% Contents : converts a (hyper)structural rule into an analytic rule
%%% Author   : Lara Katharina Spendier <lara@logic.at>

%% htar(+(Hyper)Structural, -Analytic,+StandardComplete)
%%		converts a (hyper)structural rule into an analytic rule
%%		StandardComplete: Flag if SC ( = 1) or not ( = 0)
%% 		2 steps:(1) Restructuring the current (hyper)structural rule
%%				(2) Cutting the redundant metavariables (variables occurring in the
%% 				    premises but not in the conclusion of the rule)
htar([P0, C0], [P2, C1], SC) :-
	restructuring(P0, P1, C0, C1),
%	writef('\n before cutting: SC %q \n Premises: %q',[SC, P1]),
	cutting(P1, P2,SC).
%    writef('\n after cutting: \n Premises: %q',[P2]).

%% restructuring(+Premises, -Premises, +Conclusion, -Conclusion)
%%		Premises ... premises of the rule (current + after restructuring)
%%		Conclusion ... conclusion of the rule (current + after restructuring)
restructuring(P0, P0, [], []).

restructuring(P0, P3, [[A0|['']]|T], C2) :-
	restructuring(P0, P1, T, C1),
	is_replaced_antecedent(A0, A1, P2),
	append([[A1, '']], C1, C2),
	append(P1, P2, P3).

restructuring(P0, P4, [[A0|['X'+I]]|T], C2) :-
	restructuring(P0, P1, T, C1),
	is_replaced_antecedent(A0, A1, P2),
	append(['D'+I], A1, A2),
	reverseList(A2, [], A3),
	append([[A3, 'P'+I]], C1, C2),
	append(P1, P2, P3),
	append([[['X'+I, 'D'+I], 'P'+I]], P3, P4).

%% is_replaced_antecedent(+Antecedent, -Antecedent, -Premises)
%%		Antecedent ... antecedent of the conclusion (current & after restructuring)
%%		Premises ... premises of the rule including premise for the new antecedent
is_replaced_antecedent([], [], []).

is_replaced_antecedent(['Y'+I|T], H2, P1) :-
	is_replaced_antecedent(T, H1, P0),
	append(['G'+I], H1, H2),
	append([[['G'+I], 'Y'+I]], P0, P1).

%% cutting(+Premises, -Premises, +SC)
%%		Premises ... current premises & premises after cut
%%		SC ... flag if standard complete cutting algorithm or not
cutting(P0, P0,_) :-
	is_metafree_premises(P0).
	
cutting(P0, P2,SC) :-
	\+ is_metafree_premises(P0),
	is_containing_metavariable(P0, MV),
%	writef('\n SC = %q',[SC]),
%	writef('\n      before cut_MV: Premise = %q',[P0]),
	cut_metavariable(P0, P1, MV, SC),
%	writef('\n      after cut_MV: Premise = %q',[P1]),
	cutting(P1, P2, SC).

%% cut_metavariable(+Premises, -Premises, +Metavariable, +SC)
%%		collects the premises where the metavariable occurs and cuts it
%%		Premises ... premises of the rule (current & after cut)
%%		Metavariable ... metavariable that has to be cut
%%		SC ... flag if standard complete cutting algorithm or not
%cut_metavariable(P0, P2, MV, SC) :-
cut_metavariable(P0, P2, MV, 0) :-
	collect(P0, MV, _, A, S),
	(S > 1 ->
	  sort(P0,P01)
	; P01 = P0
	),
	collect(P01, MV, P1, A, S1),
%	writef('\n          before cutp: A = %q   S = %q',[A,S1]),
%	writef('\n          before cutp: P0  = %q',[P0]),
%	writef('\n          before cutp: P01 = %q',[P01]),
%	cutp(P0, P1, P2, A, S, MV).
	cutp(P01, P1, P2, A, S1, MV, 0).

%% "simplified version" due to (com) rule for check of standard completeness
%%  Note: is only important for sequent rules, not hypersequents
cut_metavariable(P0, P2, MV, 1) :-
	collect(P0, MV, P1, A, S),
	cutp(P0, P1, P2, A, S, MV, 1).

%% collect(+Premises, +Metavariable, -PremisesMV, +NumAnte, +NumSucc)
%%		collects the parts of the premises where the metavariable occurs
%%		PremisesMV ... premises where the metavariable occurs
%%		NumAnte ... number of occurrences of the metavariable in the antecedent
%%		NumSucc ... number of occurrences of the metavariable in the succedent
collect([], _, [], 0, 0).

collect([[H, T]|T1], MV, P1, A1, S0) :-
	member(MV, H),
	collect(T1, MV, P0, A0, S0),
	append([[H, T]], P0, P1),
	A1 is A0+1.

collect([[H, MV]|T1], MV, P1, A0, S1) :-
	\+ member(MV, H),
	collect(T1, MV, P0, A0, S0),
	append([[H, MV]], P0, P1),
	S1 is S0+1.

collect([[H, T]|T1], MV, P1, A0, S0) :-
	\+ member(MV, H),
	T \= MV,
	collect(T1, MV, P0, A0, S0),
	append([], P0, P1).

%% cutp(+Premises, +PremisesMV, -Premises, +NumAnte, +NumSucc, +MV, +SC)
%%		cuts the metavariable from the premises
%%		PremisesMV ... premises where the metavariable occurs
%%		Premises ... premises (current & after the metavariable has been cut)
%%		NumAnte ... number of occurrences of the metavariable in the antecedent
%%		NumSucc ... number of occurrences of the metavariable in the succedent
%%		SC ... flag if standard complete cutting algorithm or not

cutp(_, _, _, 0, 0, _, _).

cutp(P0, PR, P1, 0, _, _, _) :-
	remove_pr(PR, P0, P1).

cutp(P0, PR, P1, _, 0, _, _) :-
	remove_pr(PR, P0, P1).

cutp(P0, PR, P3, _, NumS, MV, 0) :-
	remove_pr(PR, P0, P1),
	is_split_pr(PR, PRA, PRS, MV),
%    writef('\n          before merge: PRA = %q',[PRA]),
%    writef('\n                        PRS = %q,  NumS = %q',[PRS,NumS]),
    reduce_PRS(PRS,PRS1),
%    writef('\n                        PRS1  = %q',[PRS1]),
%	merge(PRA, PRS, P2, MV),
%	writef('\n          after merge:    P2  = %q',[P2]),
	(NumS = 1 ->
	  merge(PRA, PRS, P21, MV)
	; merge_new(PRA,PRS1,P21,NumS,MV)
	), 
%	writef('\n          after merge2:   P21 = %q',[P21]),
	%append(P1, P2, P3).
	append(P1,P21,P3).

cutp(P0, PR, P3, _, _, MV, 1) :-
	remove_pr(PR, P0, P1),
	is_split_pr(PR, PRA, PRS, MV),
	merge(PRA, PRS, P2, MV),
	append(P1, P2, P3).

%% reduce_PRS(+Premises, -ReducedPremises)
%%		reduces the premises to the antecedent
%%		Premises ... current premises 
%%		ReducedPremises ... premises reduced to the antecedent
reduce_PRS([],[]).
reduce_PRS([[A,_]|T],PRSR) :-
	reduce_PRS(T,PRSR1),
	append([A],PRSR1,PRSR).
	

%% remove_pr(+Remove, +Premises, -Premises)
%%		removes premises from the premises
%% 		Remove ... premises to be removed
%%		Premises ... current premises & after removal
remove_pr([], Premises, Premises).

remove_pr([H|T], P0, P2) :-
	remove(H, P0, P1),
	remove_pr(T, P1, P2).

%% is_split_pr(+Premises, -PremisesAnte, -PremisesSucc, +Metavariable)
%%		checks the premises and divides them in two parts: those premises where the
%%		metavariable occurs in the antecedent and those where it occurs in the succ.
%%		Premises ... current premises
%%		PremisesAnte ... premises where the metavariable occurs in the antecedent
%%		PremisesSucc ... premises where the metavariable occurs in the succedent
is_split_pr([], [], [], _).

is_split_pr([[H, MV]|T1], PRA0, PRS1, MV) :-
	is_split_pr(T1, PRA0, PRS0, MV),
	append([[H, MV]], PRS0, PRS1).

is_split_pr([[H, T]|T1], PRA1, PRS0, MV) :-
	member(MV, H),
	is_split_pr(T1, PRA0, PRS0, MV),
	append([[H, T]], PRA0, PRA1).

%% merge(+PremisesAnte, +PremisesSucc, -Premises, +Metavariable)
%%		merges the premises for the case where the metavariable occurs in both, the
%%		antecedent and the succedent
%% 		PremisesAnte ... premises where the metavariable occurs in the antecedent
%%		PremisesSucc ... premises where the metavariable occurs in the succedent
%%		Premises ... new (merged) premises
merge([], [], [], _).

merge([], _, [], _).

merge(_, [], [], _).

merge([[HA, TA]|TA1], [[HS, TS]|TS1], P2, MV) :-
	merge([[HA, TA]|TA1], TS1, PS1, MV),
	merge(TA1, [[HS, TS]], PA1, MV),
	replaceElements(HA, MV, HS, HA1),
	append([[HA1, TA]], PS1, P1),
	append(PA1, P1, P2).

%% merge_new(+PremisesAnte, +PremisesSucc, -Premises, +NumSuccedent, +Metavariable)
%%		merges the premises for the case where the metavariable occurs in both, the
%%		antecedent and the succedent
%%		--> deals with the case when there are several metavariables in one antecedent
%%		(''cut in all possible ways '')
%% 		PremisesAnte ... premises where the metavariable occurs in the antecedent
%%		PremisesSucc ... premises where the metavariable occurs in the succedent
%%		Premises ... new (merged) premises
%%		NumSuccedent ... number of premises with the metavariable in the succedent
merge_new([], [], [], _, _).

merge_new([], _, [], _, _).

merge_new(_, [], [], _, _).

%% deals with the case when there is only one premise with the metavariable
%% occurring in the succedent
merge_new([[HA, TA]|TA1], [HS1|TS1], P2, 1, MV) :-
	merge_new([[HA, TA]|TA1], TS1, PS1, 1, MV),
	merge_new(TA1, HS1, PA1, 1, MV),
	replaceElements(HA, MV, HS1, HA1),
	append([[HA1, TA]], PS1, P1),
	append(PA1, P1, P2).

%% deals with the case when there are more than 1 premises with the metavariables
%% occurring in the succedent, i.e., merging needs to be done in a combinatorial
%% way (with repetition)
merge_new([[HA, TA]|TA1], PS0, P2, NumS, MV) :-
	merge_new(TA1, PS0, PA1, NumS, MV),
	countMV(HA,MV,NumHA),
	number_premises(NumS,NumHA,NPR),
%	writef('\n === NumHA: %q === NPR: %q',[NumHA,NPR]),
	flatten(PS0,PS1),
%	writef('\n === PS1: %q',[PS1]),
	generate_comb(NumHA,PS1,0,NPR,[],Result),
%	writef('\n === Result: %q -- MV: %q',[Result,MV]),
	createNewPremises([HA, TA], Result, NewP,MV),
%	writef('\n === NewP: %q',[NewP]),
%	append([NewP],PS1,P1),
%	writef('\n === P1: %q',[P1]),
% 	replaceElements(HA, MV, HS1, HA1),
%	append([[HA1, TA]], PS1, P1),
	append(PA1, NewP, P2).

%% createNewPremises(+PremiseOrig, +PremiseGenerated, -PremiseNew, +Metavariable)
%%		merges the new premises
%% 		PremisesOrig ... original premises with variables
%%		PremisesGenerated ... premises generated with metavariables (combinatorial9
%%		PremisesNew ... new (merged) premises 
createNewPremises(_,[],[],_).
createNewPremises([Ante,Succ], [H|T], PNew, MV) :-
	createNewPremises([Ante,Succ], T, PNew1, MV),
	removeAll(MV,Ante,Ante1),
	append(H,Ante1,Ante2),
	append([[Ante2,Succ]],PNew1,PNew).

%% generate_comb(+NumA, +NumS, +Counter, +CountMax, +Existing, -Result)
%%		generates the 'cut in all possible ways' --> i.e. all possible
%%		combinations of premises with metavariables in antecedent and succedent
%% 		NumA ... Number of metavariables in antecedent
%%		NumS ... Number of premises with metavariables in the succedent
%%		Counter ... Number of already generated combinations
%%		CounterMax ... maximum of the possible generated combinations
%%		Existing ... list of combinations already generated
%%		Result ... list of all possible combinations
generate_comb(_,_,CountMax,CountMax,Result,Result). 

generate_comb(A,L,0,CountMax,[],Res) :-
	comb_with_rep(A,L,X0),
	generate_comb(A,L,1,CountMax,[X0],Res).
	
generate_comb(A,L,Count,CountMax,ResEx,Res) :-
	Count < CountMax,
	comb_with_rep(A,L,X0),
	\+ member(X0,ResEx),
	append([X0], ResEx,ResEx1),
	Count1 is Count+1,
	generate_comb(A,L,Count1,CountMax,ResEx1,Res).

generate_comb(A,L,Count,CountMax,ResEx,Res) :-
	Count < CountMax,
	comb_with_rep(A,L,X0),
	member(X0,ResEx),
	generate_comb(A,L,Count,CountMax,ResEx,Res).

%% comb_with_rep(+Counter, +List, -Result)
%%		generates the combinations (with repetition)
comb_with_rep(0,_,[]).
comb_with_rep(N,[X|T],[X|RComb]) :-
	N>0,
	N1 is N-1,
	comb_with_rep(N1,[X|T],RComb).
comb_with_rep(N,[_|T],RComb) :-
	N>0,
	comb_with_rep(N,T,RComb).

%% countMV(+List, +Metavariable, -NumberMetavariables)
%%		counts how often the metavariable occurs in a list
countMV([],_,0).
countMV([H|T],MV,NumMV) :-
  countMV(T,MV,NumMV1),
  (H = MV ->
    NumMV is NumMV1+1
  ; NumMV is NumMV1
  ).

%% number_premises(+NumS, +NumA, -NumPr)
%%		calculates how many possible combinations (with repetition) there
%%		are for the new premises
%%		formula: (n+k-1)! div (n-1)!*k!
%% 		NumS ... n = number of variables ( = succedent); Gesamtmenge
%% 		NumA ... k = number of places in antecedent; k werden ausgewählt
number_premises(NumS,NumA,NumPr) :-
  fac(NumS+NumA-1,R1),
  fac(NumS-1,R2),
  fac(NumA,R3),
  NumPr is R1/(R2*R3).

%% fac(+Number,-Factorial)
%%		calculates the factorial of Number
fac(0,1).
fac(N,F) :-
	N1 is N-1,
	fac(N1,F1),
	F is N*F1.

%% is_metafree_X (+Premises)
%%		checks if there is no metavariable in the premises
is_metafree_premises([]).

is_metafree_premises([[H,T]|T1]) :-
	is_metafree_antecedent(H),
	is_metafree_succedent(T),
	is_metafree_premises(T1).

is_metafree_antecedent([]).
is_metafree_antecedent(['G'+_|T]) :-
	is_metafree_antecedent(T).
is_metafree_antecedent(['D'+_|T]) :-
	is_metafree_antecedent(T).
is_metafree_antecedent(['P'+_|T]) :-
	is_metafree_antecedent(T).
is_metafree_antecedent([[]|T]) :-
	is_metafree_antecedent(T).
is_metafree_antecedent([''|T]) :-
	is_metafree_antecedent(T).
is_metafree_antecedent([H|T]) :-
	H \= 'Y'+_,
	H \= 'X'+_,
	\+ atomic(H),
	compound(H),
	is_metafree_antecedent(T).

is_metafree_succedent([]).
is_metafree_succedent('G'+_).
is_metafree_succedent('P'+_).
is_metafree_succedent('D'+_).
is_metafree_succedent('').
is_metafree_succedent(T) :-
	T \= 'Y'+_,
	T \= 'X'+_,
	\+ atomic(T),
	compound(T).
	
%% is_containing_metavariable(+Premises, +Metavariable)
%%		checks if the metavariable occurs in the premises
is_containing_metavariable([[[],'P'+_]|T1], MV) :-
	is_containing_metavariable(T1, MV).

is_containing_metavariable([[[],'']|T1], MV) :-
	is_containing_metavariable(T1, MV).

is_containing_metavariable([[[],T]|_],T) :-
	T \= 'P'+_,
	T \= ''.

is_containing_metavariable([[_,T]|_],T) :-
	T \= 'P'+_,
	T \= ''.

is_containing_metavariable([[[H|_],_]|_], H) :-
	H \= 'G'+_,
	H \= 'D'+_.

is_containing_metavariable([[['G'+_|T],T1]|T2],MV) :-
	is_containing_metavariable([[T,T1]|T2], MV).

is_containing_metavariable([[['D'+_|T],T1]|T2], MV) :-
	is_containing_metavariable([[T,T1]|T2], MV).

is_containing_metavariable([[[H|T],T1]|T2], MV) :-
	H \= 'X'+_,
	H \= 'Y'+_,
	\+ atomic(H),
	compound(H),
	is_containing_metavariable([[T,T1]|T2], MV).
