%%% Filename : ac_convergent.pl
%%% Date     : 21.03.2012
%%% Contents : checks whether an analytic rule in P3 is convergent
%%% Author   : Lara Katharina Spendier <lara@logic.at>


%% is_convergent(+[Prem,Con], -Convergent, -[PremR,ConvR])
%% 		[Prem,Con] ... premises and conclusion of the analytic rule
%%				that has to be checked if it is convergent
%%		Convergent ... 1 if premises are convergent, 0 otherwise
%%		[PremR,ConR]... resulting premises and conclusion 

is_convergent([Premises,Con], Convergent, [PremisesR,ConvR]) :-
	nl,
	remove_double(Premises, [], Prem),
	is_empty_RHS(Prem, ConvPrem),
	length(ConvPrem, LConvPrem),
	length(Prem, LPrem),
	( (LPrem = LConvPrem) ->
  	  Convergent = 1,
	  PremisesR = [],
	  ConvR = []
	; is_pivot0(Prem, Con, ConvPrem, ConvPrem1),
  	  length(ConvPrem1, LConvPrem1),
	  ( (LPrem = LConvPrem1) ->
  	    Convergent = 1,
	    PremisesR = [],
	    ConvR = [[0, ConvPrem1]]
	  ; 
	    ( (LConvPrem1 = LConvPrem) ->
	      Convergent = 0,
	      PremisesR = Prem,
	      ConvR = []
	    ; 
	      is_npivot_check(Prem, ConvPrem1, Check, CountPrem),
	      ( (Check = 0) ->
  	        Convergent = 0,
		PremisesR = Prem,
		ConvR = []
	      ; 
		% we need to differentiate between 0-pivot, 1-pivot, ... premises
		ConvPrem2 = [[0, ConvPrem1]],
	        recursive_pivotn(Prem, ConvPrem2, ConvPrem3, CountPrem, 1),
	        is_convergent_length(ConvPrem3, LConvPrem3),
	        ( (LPrem = LConvPrem3) ->
	           Convergent = 1
	        ;  Convergent = 0
  	        ),
		PremisesR = Prem,
		ConvR = ConvPrem3
     	      )
	    )
	  )
	).

%% is_empty_RHS(+Premise, -ConvPremise)
%% 	checks condition 1: a premise is convergent if it has an empty RHS
%%	Premise     ... Premises of the rule
%%	ConvPremise ... convergent premises (with an empty RHS)
is_empty_RHS([],[]).
is_empty_RHS([H|T], Empty) :-
	H = [_, ''],
	is_empty_RHS(T,Empty1),
	append([H],Empty1,Empty).
is_empty_RHS([H|T], Empty) :-
	H \= [_, ''],
	is_empty_RHS(T,Empty1),
	Empty=Empty1.

%% is_pivot0(+Premise, +Conclusion, +ConvergentPremises, -ConvergentPremises)
%% 	checks condition 2: a premise is convergent if it is a 0-pivot
%%	i.e., there exists one component in the conclusion that has the exact same RHS 
%% 	and all the metavariables of the LHS of the premise are contained in the LHS
%%	of the component in the conclusion
%%	ConvergentPremises ... the calculated convergent premises are extended by
%%			       the 0-pivot premises
is_pivot0([], _, ConvPrem, ConvPrem).
is_pivot0([H|T], Con, ConvPrem, ConvPrem1) :-
	member(H, ConvPrem),
	is_pivot0(T, Con, ConvPrem, ConvPrem1).
is_pivot0([H|T], Con, ConvPrem, ConvPrem1) :-
	\+ member(H, ConvPrem),
	check_pivot_RHS(H, Con, ConvPrem2),
	( (ConvPrem2 \= []) -> 
	  append([ConvPrem2], ConvPrem, ConvPrem3)
	; ConvPrem3 = ConvPrem
	),
	is_pivot0(T, Con, ConvPrem3, ConvPrem1).

%% check_pivot_RHS(+Premise, +Conclusion, -PivotPremises)
%%	checks if the RHS of the premise is the same as the RHS of the conclusion
%%	if yes, we check the LHS of the premise (if it is the same as the LHS of the conclusion)
%%	otherwise, we look at the next component of the conclusion
check_pivot_RHS(_, [], []).

check_pivot_RHS([PLHS, PRHS], [[CLHS,PRHS]|_], PivotPrem):-
	check_pivot_LHS(PLHS, CLHS, PivotPrem, [PLHS, PRHS]).

check_pivot_RHS([PLHS, PRHS], [[_,RHS]|T], PivotPrem):-
	PRHS \= RHS,
	check_pivot_RHS([PLHS, PRHS], T, PivotPrem).

%% check_pivot_LHS(+LHSPremise, +LHSConclusion, -PivotPremises, +Premise)
%%	checks if the left hand side of a premise is contained in the LHS
%%	of one component of the conclusion
%%	+LHSPremise    ... LHS of the premise
%%	+LHSConclusion ... LHS of one component of the conclusion
%%	+Premise 	... the premise (LHS+RHS)
check_pivot_LHS([], _, Prem, Prem).

check_pivot_LHS([PH|PT], LHS, PivotPrem, Prem) :-
	member(PH, LHS),
	check_pivot_LHS(PT, LHS, PivotPrem, Prem).

check_pivot_LHS([PH|_], LHS, PivotPrem, _) :-
	\+ member(PH, LHS),
	PivotPrem = [].

%% recursive_pivotn (+Premises, +ConvergentPrem, -ConvergentPremR, +Count, +Pivot)
%%	we start to recursively calculate the 1-pivot, ..., n-pivot premises and either
%%	stop when there are no premises any more, or we have checked all premises
%%	Premises ...	The premises we are starting with (decreasing!)
%%	ConvPrem ...	input (and output) list containing convergent premises
%%	Pivot	...	We are looking for the Pivot premises, start with 1-pivot premises
recursive_pivotn([], ConvPrem, ConvPrem, _, _).
recursive_pivotn(_, ConvPrem, ConvPrem, 0, _).
recursive_pivotn(Prem, ConvPrem, ConvPremR, _, 1) :-
	ConvPrem = [[0, Conv0]],
	check_pivotn0(Prem, Prem1, Conv0),
	check_pivotn(Prem1, Prem2, Conv0, Conv1, 1),
	append([[1, Conv1]], ConvPrem, ConvPrem1),
	length(Prem2, CountPrem2),
	recursive_pivotn(Prem2, ConvPrem1, ConvPremR, CountPrem2, 2).

recursive_pivotn(Prem, ConvPrem, ConvPremR, _, N) :-
	NP is N-1,
	member([NP, ConvN], ConvPrem),
	check_pivotn(Prem, Prem1, ConvN, Conv1, N),
	( (Conv1 \= []) ->
	  append([[N, Conv1]], ConvPrem, ConvPrem1),
 	  length(Prem1, CountPrem1),
	  N1 is N+1,
 	  recursive_pivotn(Prem1, ConvPrem1, ConvPremR, CountPrem1, N1)
	; ConvPremR = ConvPrem
	).

%% is_convergent_length (+PivotPremises, -Length)
%%	We calculate the number of x-pivot premises: [x, [Prem]]
%%	PivotPremises ... n-pivot premises of the form: [n, [[[],_], ...]
%%	Length	      ... number of n-pivot premises
is_convergent_length([], 0).
is_convergent_length([[_, H]|T], Length) :-
	length(H, LengthH),
	is_convergent_length(T, LengthT),
	Length is LengthH + LengthT.
	

%% check_pivotn0(+Premises, -Premises (without 0-pivot), +Convergent)
%%	we skip the 0-pivot premises and delete them from the premises-list
check_pivotn0([], [], _).
check_pivotn0(Prem, PremR, Conv0) :-
	Prem = [H|T],
	member(H, Conv0),
	check_pivotn0(T, PremR, Conv0).
check_pivotn0(Prem, PremR, Conv0) :-
	Prem = [H|T],
	\+ member(H, Conv0),
	check_pivotn0(T, Prem1, Conv0),
	append([H], Prem1, PremR).	

%% check_pivotn(+Premises, -Premises, +Convergent, -Convergent, +PivotCount)
%%	we check if we have 1-/2-/.../n-pivot premises recursively for all premises
check_pivotn([], [], _, [], _).
check_pivotn([H|T], PremR, Conv, ConvR, PC) :-
	is_pivotn(H, Conv, PC, Check, 0),
	check_pivotn(T, Prem1, Conv, Conv1, PC),
	( (Check is 0) ->
	  append([H], Prem1, PremR),
	  ConvR = Conv1
	; PremR = Prem1,
	  append([H], Conv1, ConvR)
	).


%% is_pivotn(+Prem, +ConvPrem, PivotCount, Check, PPCount),
%% 	we check if the actual premise (+Prem) is an n-pivot (PivotCount) premise
%%	return Check=0 if it is not an n-pivot (PivotCount) premise, Check=1 otherwise
is_pivotn(_, [], _, 0, _).

is_pivotn(Prem, Conv, _, Check, _) :-
	member(Prem, Conv),
	Check = 0.

is_pivotn(Prem, Conv, PC, Check, PPCount) :-
	\+ member(Prem, Conv),
	Conv = [H|T],
	Prem = [_, S],
	H \= [_, S],
	is_pivotn(Prem, T, PC, Check, PPCount).

is_pivotn(Prem, Conv, PC, Check, PPCount) :-
	\+ member(Prem, Conv),
	Conv = [H|T],
	check_pivot_differences(Prem, H, PC, Check1),
	( (Check1 is 1) ->
	  PPCount1 is PPCount + 1
	; PPCount1 = PPCount
	),
	( (PPCount1 is PC) ->
	  Check = 1
	; is_pivotn(Prem, T, PC, Check, PPCount1)
	).

%% check_pivot_differences(+Prem, +ConvPrem, PivotCount, -Check)
%% 	actual calculation for the n-pivot
%%	* for every n-pivot there must exist n (n-1)-pivot premises
%%	=> we have to count bottom-up (starting with the 0-pivot, ...) for the
%%	  premise to check
check_pivot_differences(_, [], _, 0).
check_pivot_differences(Prem, Conv, _, Check) :-
	Prem = [A, S],
	Conv = [AP0, S],
	is_LHS_difference(A, AP0, DiffL, DiffP0L1, _),
	remove_double(DiffL, [], DiffL1),
	length(DiffL1, Diff1),
	length(DiffP0L, DiffP0),
	remove_double(DiffP0L, [], DiffP0L2),
	length(DiffP0L2, DiffP02),
	( ((DiffP01 = Diff1; DiffP0 = Diff1), Diff1 is 1) ->
%	( (DiffP01 = Diff1, Diff1 is 1) ->
 	  Check is 1
	; Check is 0
	).

%% is_LHS_difference
%%	checks the actual differences between the 0-pivot and the current premise
is_LHS_difference([], P0L, [], P0L, 0).
is_LHS_difference([H|T], P0L, DiffL, DiffP0L, Diff) :-
	member(H, P0L),
	remove_once([H], P0L, P0L1),
	is_LHS_difference(T, P0L1, DiffL1, DiffP0L, Diff1),
	DiffL = DiffL1, 
	Diff is Diff1.
is_LHS_difference([H|T], P0L, DiffL, DiffP0L, Diff) :-
	\+ member(H, P0L),
	is_LHS_difference(T, P0L, DiffL1, DiffP0L, Diff1),
	Diff is Diff1 + 1,
	append([H], DiffL1, DiffL).


is_npivot_check([], _, 1, 0).
is_npivot_check([H|T], ConvPrem, Check, CountPrem) :-
	member(H, ConvPrem),
	is_npivot_check(T, ConvPrem, Check, CountPrem).
is_npivot_check([H|T], ConvPrem, Check, CountPrem) :-
	\+ member(H, ConvPrem),
	H = [A, S],
	length(A, LPrem),
	is_0pivot_length(S, LPrem, ConvPrem, Check1),
	( (Check1 = 1) ->
	  is_npivot_check(T, ConvPrem, Check, CountPrem1),
	  CountPrem is CountPrem1 + 1
	; Check = 0, CountPrem = 0
	).

is_0pivot_length(_, _, [], 0).
is_0pivot_length(S, LPrem, [[_, '']|T], Check) :-
	is_0pivot_length(S, LPrem, T, Check).
is_0pivot_length(S, LPrem, [[H, _]|T], Check) :-
	length(H, LPP),
	LPrem \= LPP,
	is_0pivot_length(S, LPrem, T, Check).
is_0pivot_length(S, LPrem, [[H, S1]|T], Check) :-
	length(H, LPP),
	LPrem = LPP,
	S \= S1,
	is_0pivot_length(S, LPrem, T, Check).
is_0pivot_length(S, LPrem, [[H, S1]|_], Check) :-
	length(H, LPP),
	LPrem = LPP,
	S = S1,
	Check = 1.

