%%% Filename : ac_hyperstructural.pl
%%% Date     : 23.02.2011
%%% Contents : converts an axiom into a (hyper)structural rule
%%% Author   : Lara Katharina Spendier <lara@logic.at>

%% athsr(+Axiom, +Class, -Structural)
%%		creates the (hyper)structural rule from an axiom
athsr(Axiom, C, Structural) :-
	member(C, [p(1), n(1), n(2)]),
	apply_to_conclusion(_, P1, [Axiom], C1),
	is_equivalence_left(P1, P2, C1, C2, _),
	apply_to_premises(P2, P3),
	is_equivalence_right(P3, P4, C2, C3, _),
	apply_to_premises_left(P4, P5),
	remove([[],''], P5, P6),
	Structural = [P6, C3].

athsr(HS, C, Structural) :-
	member(C, [p(2), p(3)]),
	apply_to_conclusion(_, P1, HS, C1),
	is_equivalence_left(P1, P2, C1, C2, _),
	apply_to_premises(P2, P3),
	is_equivalence_right(P3, P4, C2, C3, _),
	apply_to_premises_left(P4, P5),
	remove([[],''], P5, P6),
	Structural = [P6, C3].
	
%% is_hypersequent(+Axiom, -Hypersequent)
%% 		creates a hypersequent from a P3-axiom (an N2-axiom is a hypersequent with only
%% 		one component)
is_hypersequent(Ax1 v Ax2, HS1) :-
	is_hypersequent(Ax1, HS0),
	append([Ax2], HS0, HS1).

is_hypersequent(Ax, HS) :-
	atomic(Ax),
	append([Ax], [], HS).

is_hypersequent(Ax, HS) :-
	compound(Ax),
	append([Ax], [], HS).

%% is_several_hyps(+Hypersequent, -Hypersequent)
%% 		checks whether the hypersequent contains applications of & and, if yes,
%%		creates more hypersequents
is_several_hyps([],[]).
is_several_hyps([H|T], New) :-
	member((F1 & F2), H),
	remove((F1 & F2), H, H1),
	append([F1],H1,H2),
	append([F2],H1,H3),
	append([H3],[H2],H4),
	is_several_hyps(H4, New1),
	is_several_hyps(T, New2),
	( New2 = [] ->
	  New = New1
	; append(New2,New1,New)
	).
is_several_hyps([H|T], New) :-
	\+ member((_ & _), H),
	is_several_hyps(T, New1),
	append([H],New1,New).
	

%% apply_to_conclusion (+Antecedent, -Antecedent, +Succedent, -Succedent)
%%		apply the invertible rules (first right, then left) to the
%%		conclusion => step (1)
apply_to_conclusion(_, [], [], []).

apply_to_conclusion(P0, P4, [H|T], C4) :-
	apply_to_conclusion(_, P1, T, C1),
	apply_right(P0, P2, H, C2),
	apply_left(P2, P3, C2, C3),     
	append([C3], C1, C4),
	append(P3, P1, P4).
	
%% is_equivalence_right / is_equivalence_left (+Premises, -Premises, +Conclusion, -Conclusion, _)
%%		applies the left/right equivalence rules to the rule
is_equivalence_right(_, [], [], [], 1).
is_equivalence_right(P0, P3, [H|T], C3, NR1) :-
	is_equivalence_right(_, P1, T, C1, NR0),
	apply_equivalence_right(P0, P2, H, C2, NR0),	
	NR1 is NR0+1,
	append([C2], C1, C3),
	append(P2, P1, P3).

is_equivalence_left(_, [], [], [], 1).
is_equivalence_left(P0, P3, [H|T], C3, NL2) :-
	is_equivalence_left(_, P1, T, C1, NL0),
	apply_equivalence_left(P0, P2, H, C2, NL0, NL1),
	NL2 is NL0+NL1,
	append([C2], C1, C3),
	append(P2, P1, P3).

%% apply_right (+Premises, -Premises, +SuccedentConclusion, -Conclusion)
%%		apply the invertible right rules
% (-> r) rule: X |- a -> b => a, X |- b
apply_right(PR0, PR1, Ax1 -> Ax2, [P1, C0]) :-
	apply_right(PR0, PR1, Ax2, [P0, C0]),
	append(P0, [Ax1], P1).
	
% (-r) rule: X |- -a => a, X |-
apply_right([], [], -C1, [[C1], '']).
	
apply_right([], [], Axiom, [[], Axiom]).


%% apply_right_land (+Premises, -Premises, +Conclusion, -Conclusion)
%%		apply the logical-and invertible right rule
% (&r) rule: X |- a & b => X |- a  X |- b / X |- a & b
apply_right_land(P0, P1, [P, Ax1 & Ax2], [P, Ax1 & Ax2]) :-
	write('apply right land'),
	append([[P, Ax1], [P, Ax2]], P0, P1).

apply_right_land(P0, P0, [P, C], [P, C]).

%% apply_left (+Premises, -Premises, +Conclusion, -AntecedentConclusion)
%%		applies the invertible left rules
% (*l)-rule: A*A, X |- Y => A, A, X |- Y
apply_left(P0, P2, [P, C], A) :-
	member(Ax1*Ax2, P),
	remove(Ax1*Ax2, P, P1),
	Ax3 = [Ax1,Ax2],
	append(Ax3, P1, Ax4),
	apply_left(P0, P2, [Ax4, C], A).

% (1l)-rule: 1, X |- Y => X |- Y
apply_left(P0, P2,[P, C], A) :-
	member(1, P),
	remove(1, P, P1),
	apply_left(P0, P2, [P1, C], A).
	
apply_left(P0, P1, [Axiom, C], [Axiom, C]) :-
	atomic(Axiom),
	append([], P0, P1).

apply_left(P0, P1, [Axiom, C], [Axiom, C]) :-
	compound(Axiom),
	append([], P0, P1).

%% apply_equivalence_right(+Premises, -Premises, +Conclusion, -Conclusion, _)
%%		applies the equivalence rule for the succedent
apply_equivalence_right(Premises, Premises, [P, ''], [P, ''], _).

apply_equivalence_right(Premises, Premises, [P, 0], [P, ''], _).

apply_equivalence_right(Premises, Premises, [P, 'X'+I], [P, 'X'+I], _).

apply_equivalence_right(P0, P1, [P, C], [P, 'X'+NR], NR) :-
	append([[[C], 'X'+NR]], P0, P1).
	
%% apply_equivalence_left(+Premises, -Premises, +Conclusion, -Conclusion, _)
%%		applies the equivalence rule for the antecedent
apply_equivalence_left(PR0, PR2, [P0, C], [P1, C], NL, NL1) :-
	are_premises(P0, P1, PR1, NL, NL1),
	append(PR1, PR0, PR2).

%% are_premises (+AntecedentConclusion, -AntecedentConclusion, -Premises, _, _ )
%%		creates the premises when applying the equivalence-left rule
%%		AntecedentConclusion ... antecedent of the conclusion of the rule
%%		Premises ... premises of the rule (at least partially) extended with the antecedent
are_premises([], [], [], _, 0).

are_premises([H|T], PR1, P1, N0, NA1) :-
	N1 is N0+1,
	are_premises(T, PR0, P0, N1, NA0),
	NA1 is NA0+1,
	append(['Y'+N0], PR0, PR1),
	append([[['Y'+N0], H]], P0, P1).

%% apply_to_premises(_left) (+Premises, -Premises)
%%		applies the invertible rules (left & right) to the premises after application
%%		of the equivalence rules => step (3)
apply_to_premises(P0, P0) :-
	is_pr_left_atomic(P0),
	is_pr_right_atomic(P0).

apply_to_premises(P0, P3) :-
	\+ is_pr_left_atomic(P0),
	\+ is_pr_right_atomic(P0),
	apply_left_premises(P0, P1),
	apply_right_premises(P1, P2),
	apply_to_premises(P2, P3).

apply_to_premises(P0, P2) :-
	\+ is_pr_left_atomic(P0),
	is_pr_right_atomic(P0),
	apply_left_premises(P0, P1),
	apply_to_premises(P1, P2).

apply_to_premises(P0, P2) :-
	is_pr_left_atomic(P0),
	\+ is_pr_right_atomic(P0),
	apply_right_premises(P0, P1),
	apply_to_premises(P1, P2).

apply_to_premises_left(P0, P0) :-
	is_pr_left_atomic(P0).
	
apply_to_premises_left(P0, P2) :-
	\+ is_pr_left_atomic(P0),
	apply_left_premises2(P0, P1),
	apply_to_premises_left(P1, P2).

%% is_pr_left_atomic(+Premises), is_pr_right_atomic(+Premises)
%%		checks whether the premises are left/right atomic, which means that the 
%%		application of left/right invertible rules is not possible
is_pr_left_atomic([]).
is_pr_left_atomic([[H,_]|T]) :-
	\+ member((_ v _), H),
	\+ member((_ * _), H),
	\+ member(1, H),
	\+ member([1], H),
	is_pr_left_atomic(T).

is_pr_right_atomic([]).
is_pr_right_atomic([[_,T1]|T]) :-
	T1 \= (_ -> _),
	T1 \= -(_),
	T1 \= (_ & _),
	T1 \= 0,
	is_pr_right_atomic(T).

%% apply_left_premises (+Premises, -Premises)
%% apply_left_premises2 (+Premises, -Premises)
%%		applies the left invertible rules to the premises
%%		difference between apply_left_premises and apply_left_premises2: the latter also
%%		applies (vl) whereas the first one does not
apply_left_premises([], []).

apply_left_premises([[H,T1]|T], P2) :-
	T1 \= 'X'+_,
	member((A1 * A2), H),
	apply_left_premises(T, P1),
	remove((_ * _), H, H1),
	append([A1, A2], H1, H2),
	append([[H2, T1]], P1, P2).

apply_left_premises([[H,T1]|T], P2) :-
	T1 \= 'X'+_,
	member(1, H),
	apply_left_premises(T, P1),
	remove(1, H, H1),
	append([[H1, T1]], P1, P2).

apply_left_premises([[H,T1]|T], P2) :-
	T1 \= 'X'+_,
	apply_left_premises(T, P1),
	append([[H, T1]], P1, P2).
	
apply_left_premises([[[A1 * A2],'X'+I]|T], P2) :-
	apply_left_premises(T, P1),
	append([[[A1, A2], 'X'+I]], P1, P2).

apply_left_premises([[[1],'X'+I]|T], P2) :-
	apply_left_premises(T, P1),
	append([[[], 'X'+I]], P1, P2).
	
apply_left_premises([[H,'X'+I]|T], P2) :-
	atomic(H),
	apply_left_premises(T, P1),
	append([[[H], 'X'+I]], P1, P2).

apply_left_premises([[H,'X'+I]|T], P2) :-
	compound(H),
	apply_left_premises(T, P1),
	append([[H, 'X'+I]], P1, P2).

apply_left_premises2([], []).

apply_left_premises2([[H,T1]|T], P3) :-
	T1 \= 'X'+_,
	member((A1 v A2), H),
	apply_left_premises2(T, P1),
	remove((_ v _), H, H1),
	append([A1], H1, HA1),
	append([A2], H1, HA2),
	append([[HA1, T1]], P1, P2),
	append([[HA2, T1]], P2, P3).

apply_left_premises2([[[A1 v A2],'X'+I]|T], P3) :-
	apply_left_premises2(T, P1),
	append([[[A1], 'X'+I]], P1, P2),
	append([[[A2], 'X'+I]], P2, P3).

apply_left_premises2([[H,T1]|T], P2) :-
	member((A1 * A2), H),
	apply_left_premises2(T, P1),
	remove((A1 * A2), H, H1),
	append([A1, A2], H1, H2),
	append([[H2, T1]], P1, P2).

apply_left_premises2([[H,T1]|T], P2) :-
	member(1, H),
	apply_left_premises2(T, P1),
	remove(1, H, H1),
	append([[H1, T1]], P1, P2).

apply_left_premises2([[H,T1]|T], P2) :-
	apply_left_premises2(T, P1),
	append([[H, T1]], P1, P2).
	
apply_left_premises2([[[1],'X'+I]|T], P2) :-
	apply_left_premises2(T, P1),
	append([[[], 'X'+I]], P1, P2).
	
apply_left_premises2([[H,'X'+I]|T], P2) :-
	atomic(H),
	apply_left_premises2(T, P1),
	append([[[H], 'X'+I]], P1, P2).

apply_left_premises2([[H,'X'+I]|T], P2) :-
	compound(H),
	apply_left_premises2(T, P1),
	append([[H, 'X'+I]], P1, P2).

%% apply_right_premises (+Premises, -Premises)
%%		applies the right invertible rules to the premises
apply_right_premises([], []).

apply_right_premises([[H,'X'+I]|T], P2) :-
	apply_right_premises(T, P1),
	append([[H, 'X'+I]], P1, P2).
	
apply_right_premises([[H,A1 -> A2]|T], P2) :-
	apply_right_premises(T, P1),
	append([A1], H, H1),
	append([[H1, A2]], P1, P2).
	
apply_right_premises([[H,-(A)]|T], P2) :-
	apply_right_premises(T, P1),
	append([A], H, H1),
	append([[H1, '']], P1, P2).

apply_right_premises([[H,0]|T], P2) :-
	apply_right_premises(T, P1),
	append([[H, '']], P1, P2).
	
apply_right_premises([[H,A1 & A2]|T], P3) :-
	apply_right_premises(T, P1),
	append([[H, A1]], P1, P2),
	append([[H, A2]], P2, P3).	

apply_right_premises([[H,T1]|T], P2) :-
	atomic(T1),
	apply_right_premises(T, P1),
	append([[H, T1]], P1, P2).

